import csv
import glob
import json
import os

import pandas as pd
from typing import Dict, List, Tuple, Set

from config import TEMPORARY_OUTPUT_PATH
from study1.s1_config import BACKGROUND_LOOKUP, INPUT_PATH, DATA_SET_INPUT_PATH, QUESTIONNAIRE_INPUT_PATH
from utils import data_string_converter


def get_visualization_data() -> Dict[Tuple[int, int], Dict[str, List[int]]]:
    output = {}

    for fp in glob.glob(f"{DATA_SET_INPUT_PATH}/*-data-linechart.csv"):
        file_name = fp.split('\\')[-1]
        file_nr = int(file_name.split('\\')[-1][0:3])
        block_nr = file_nr % 5 if file_nr > 0 else file_nr
        block_nr = 5 if block_nr == 0 else block_nr
        vis_nr = int((file_nr - block_nr) / 5 + 1) if file_nr > 0 else 1

        file_dict = {}
        with open(fp, newline="") as csv_file:
            reader = csv.reader(csv_file, delimiter=',')
            for i, row in enumerate(reader):
                if i == 0:
                    continue

                if row[2] not in file_dict:
                    file_dict[row[2]] = []
                file_dict[row[2]].append(int(row[1]))

        output[(block_nr, vis_nr)] = file_dict

    return output


def calc_correct_answers(reload_csv: bool = False) -> pd.DataFrame:
    os.makedirs(TEMPORARY_OUTPUT_PATH, exist_ok=True)
    file_path = f"{TEMPORARY_OUTPUT_PATH}/S1-correct-answer-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        df = pd.read_csv(file_path)
        df["value"] = df.apply(lambda row: data_string_converter(row["value"]), axis=1)
        return df

    vis_data = get_visualization_data()
    block_order = {
        1: 1, 6: 1, 11: 1, 16: 1, 21: 1,
        2: 2, 7: 2, 12: 2, 17: 2, 22: 2,
        3: 3, 8: 3, 13: 3, 18: 3, 23: 3,
        4: 4, 9: 4, 14: 4, 19: 4, 24: 4
    }
    pre_output = {(b, q + 1): [bt, b, q + 1] for b, bt in block_order.items() for q in range(0, 3 if b % 5 == 3 else 4)}

    # Block 1
    v2, v1 = find_extreme_value(vis_data[(1, 1)], "max")
    v4, v3 = find_extreme_value(vis_data[(1, 1)], "min", 4)
    v1, v3 = list(v1), list(v3)
    pre_output[(1, 1)].append(str(v1))
    pre_output[(1, 2)].append(v2)
    # pre_output[(1, 3)].append(f"{v3} | {v4}")
    pre_output[(1, 3)].append(v3)
    pre_output[(1, 4)].append(abs(v4 - v2))
    v2, v1 = find_extreme_value(vis_data[(1, 2)], "min")
    v4, v3 = find_extreme_value(vis_data[(1, 2)], "max", 4)
    v1, v3 = list(v1), list(v3)
    pre_output[(6, 1)].append(str(v1))
    pre_output[(6, 2)].append(v2)
    # pre_output[(6, 3)].append(f"{v3} | {v4}")
    pre_output[(6, 3)].append(v3)
    pre_output[(6, 4)].append(abs(v4 - v2))
    v2, v1 = find_extreme_value(vis_data[(1, 3)], "max")
    v4, v3 = find_extreme_value(vis_data[(1, 3)], "max", 5)
    v1, v3 = list(v1), list(v3)
    pre_output[(11, 1)].append(str(v1))
    pre_output[(11, 2)].append(v2)
    # pre_output[(11, 3)].append(f"{v3} | {v4}")
    pre_output[(11, 3)].append(v3)
    pre_output[(11, 4)].append(abs(v4 - v2))
    v2, v1 = find_extreme_value(vis_data[(1, 4)], "min")
    v4, v3 = find_extreme_value(vis_data[(1, 4)], "min", 5)
    v1, v3 = list(v1), list(v3)
    pre_output[(16, 1)].append(str(v1))
    pre_output[(16, 2)].append(v2)
    # pre_output[(16, 3)].append(f"{v3} | {v4}")
    pre_output[(16, 3)].append(v3)
    pre_output[(16, 4)].append(abs(v4 - v2))
    v2, v1 = find_extreme_value(vis_data[(1, 5)], "min")
    v4, v3 = find_extreme_value(vis_data[(1, 5)], "max", 5)
    v1, v3 = list(v1), list(v3)
    pre_output[(21, 1)].append(str(v1))
    pre_output[(21, 2)].append(v2)
    # pre_output[(21, 3)].append(f"{v3} | {v4}")
    pre_output[(21, 3)].append(v3)
    pre_output[(21, 4)].append(abs(v4 - v2))
    # Block 2
    pre_output[(2, 1)].append(6)  # Line 3
    pre_output[(2, 2)].append(35)
    pre_output[(2, 3)].append(16)  # x = 1
    pre_output[(2, 4)].append(len(find_extreme_lines(vis_data[(2, 1)], "max", 16, 1)))
    pre_output[(7, 1)].append(9)  # Line 4
    pre_output[(7, 2)].append(3)
    pre_output[(7, 3)].append(35)  # x = 8
    pre_output[(7, 4)].append(len(find_extreme_lines(vis_data[(2, 2)], "min", 35, 1)))
    pre_output[(12, 1)].append(6)  # Line 3
    pre_output[(12, 2)].append(34)
    pre_output[(12, 3)].append(4)  # x = 1, 5, 8
    pre_output[(12, 4)].append(len(find_extreme_lines(vis_data[(2, 3)], "min", 4, 1)))
    pre_output[(17, 1)].append(9)  # Line 0
    pre_output[(17, 2)].append(4)
    pre_output[(17, 3)].append(24)  # x = 4, 6
    pre_output[(17, 4)].append(len(find_extreme_lines(vis_data[(2, 4)], "min", 24, 1)))
    pre_output[(22, 1)].append(5)  # Line 0
    pre_output[(22, 2)].append(9)
    pre_output[(22, 3)].append(37)  # x = 7, 9
    pre_output[(22, 4)].append(len(find_extreme_lines(vis_data[(2, 5)], "min", 37, 1)))
    # Block 3
    v1 = find_extreme_lines(vis_data[(3, 1)], "max", 45, 1)
    pre_output[(3, 1)].append(len(v1))
    v2 = f"{find_extreme_value({k: v for k, v in vis_data[(3, 1)].items() if k in v1}, 'min', 1)[0]}-{find_extreme_value({k: v for k, v in vis_data[(3, 1)].items() if k in v1}, 'max', 1)[0]}"
    pre_output[(3, 2)].append(v2)
    pre_output[(3, 3)].append(get_average({k: v for k, v in vis_data[(3, 1)].items() if k in v1}, list(range(0, 8)), "total")["total"])
    v1 = find_extreme_lines(vis_data[(3, 2)], "max", 30, 1)
    pre_output[(8, 1)].append(len(v1))
    v2 = f"{find_extreme_value({k: v for k, v in vis_data[(3, 2)].items() if k in v1}, 'min', 1)[0]}-{find_extreme_value({k: v for k, v in vis_data[(3, 2)].items() if k in v1}, 'max', 1)[0]}"
    pre_output[(8, 2)].append(v2)
    pre_output[(8, 3)].append(get_average({k: v for k, v in vis_data[(3, 2)].items() if k in v1}, list(range(0, 10)), "total")["total"])
    v1 = find_extreme_lines(vis_data[(3, 3)], "min", 15, 1)
    pre_output[(13, 1)].append(len(v1))
    v2 = f"{find_extreme_value({k: v for k, v in vis_data[(3, 3)].items() if k in v1}, 'min', 1)[0]}-{find_extreme_value({k: v for k, v in vis_data[(3, 3)].items() if k in v1}, 'max', 1)[0]}"
    pre_output[(13, 2)].append(v2)
    pre_output[(13, 3)].append(get_average({k: v for k, v in vis_data[(3, 3)].items() if k in v1}, list(range(0, 12)), "total")["total"])
    v1= find_extreme_lines(vis_data[(3, 4)], "min", 5, 1)
    pre_output[(18, 1)].append(len(v1))
    v2 = f"{find_extreme_value({k: v for k, v in vis_data[(3, 4)].items() if k in v1}, 'min', 1)[0]}-{find_extreme_value({k: v for k, v in vis_data[(3, 4)].items() if k in v1}, 'max', 1)[0]}"
    pre_output[(18, 2)].append(v2)
    pre_output[(18, 3)].append(get_average({k: v for k, v in vis_data[(3, 4)].items() if k in v1}, list(range(0, 10)), "total")["total"])
    v1 = find_extreme_lines(vis_data[(3, 5)], "max", 25, 1)
    pre_output[(23, 1)].append(len(v1))
    v2 = f"{find_extreme_value({k: v for k, v in vis_data[(3, 5)].items() if k in v1}, 'min', 1)[0]}-{find_extreme_value({k: v for k, v in vis_data[(3, 5)].items() if k in v1}, 'max', 1)[0]}"
    pre_output[(23, 2)].append(v2)
    pre_output[(23, 3)].append(get_average({k: v for k, v in vis_data[(3, 5)].items() if k in v1}, list(range(0, 10)), "total")["total"])
    # Block 4
    v1 = find_extreme_lines(vis_data[(4, 1)], "min", 35, 3)
    pre_output[(4, 1)].append(len(v1))
    v3, v2 = find_extreme_value({k: v for k, v in vis_data[(4, 1)].items() if k in v1}, "max", 1)
    v2 = list(v2)
    pre_output[(4, 2)].append(str(v2))
    pre_output[(4, 3)].append(v3)
    pre_output[(4, 4)].append(json.dumps(get_average(vis_data[(4, 1)], list(range(0, 8)), "perX")))
    v1 = find_extreme_lines(vis_data[(4, 2)], "max", 35, 3)
    pre_output[(9, 1)].append(len(v1))
    v3, v2 = find_extreme_value({k: v for k, v in vis_data[(4, 2)].items() if k in v1}, "min", 1)
    v2 = list(v2)
    pre_output[(9, 2)].append(str(v2))
    pre_output[(9, 3)].append(v3)
    pre_output[(9, 4)].append(json.dumps(get_average(vis_data[(4, 2)], list(range(0, 10)), "perX")))
    v1 = find_extreme_lines(vis_data[(4, 3)], "max", 34, 3)
    pre_output[(14, 1)].append(len(v1))
    v3, v2 = find_extreme_value({k: v for k, v in vis_data[(4, 3)].items() if k in v1}, "min", 1)
    v2 = list(v2)
    pre_output[(14, 2)].append(str(v2))
    pre_output[(14, 3)].append(v3)
    pre_output[(14, 4)].append(json.dumps(get_average(vis_data[(4, 3)], list(range(0, 12)), "perX")))
    v1 = find_extreme_lines(vis_data[(4, 4)], "max", 33, 3)
    pre_output[(19, 1)].append(len(v1))
    v3, v2 = find_extreme_value({k: v for k, v in vis_data[(4, 4)].items() if k in v1}, "min", 1)
    v2 = list(v2)
    pre_output[(19, 2)].append(str(v2))
    pre_output[(19, 3)].append(v3)
    pre_output[(19, 4)].append(json.dumps(get_average(vis_data[(4, 4)], list(range(0, 10)), "perX")))
    v1 = find_extreme_lines(vis_data[(4, 5)], "max", 25, 3)
    pre_output[(24, 1)].append(len(v1))
    v3, v2 = find_extreme_value({k: v for k, v in vis_data[(4, 5)].items() if k in v1}, "min", 1)
    v2 = list(v2)
    pre_output[(24, 2)].append(str(v2))
    pre_output[(24, 3)].append(v3)
    pre_output[(24, 4)].append(json.dumps(get_average(vis_data[(4, 5)], list(range(0, 10)), "perX")))

    output = []
    for row in pre_output.values():
        # Save the values to the dict
        value = row[-1]
        # Add additional information
        # if row[0] is 1 and row[2] is 3:
        #     value, value_y = value.split('|')
        #     output.append(row[:len(row) - 1] + [float(value_y), "Y-Value"])
        if row[0] is 3 and row[2] is 2:
            min_v, max_v = value.split('-')
            output.append(row[:len(row) - 1] + [float(min_v), "Min"])
            output.append(row[:len(row) - 1] + [float(max_v), "Max"])
            output.append(row[:len(row) - 1] + [abs(float(min_v) - float(max_v)), "Delta"])
        if type(value) is str and ("[" in value or "{" in value) and "|" not in value:
            output.append(row[:len(row) - 1] + [json.loads(value), ""])
        else:
            output.append(row + [""])

    output_dict = {
        "blockType": [v[0] for v in output],
        "block": [v[1] for v in output],
        "complexity": [int(v[1] / 5 + 1) for v in output],
        "questionNr": [v[2] for v in output],
        "note": [v[4] for v in output],
        "value": [v[3] for v in output]
    }
    output = pd.DataFrame(output_dict)

    output.to_csv(file_path)
    return pd.DataFrame(output)


def find_extreme_value(lines: Dict[str, List[int]], extreme: str, nr: int = 1) -> Tuple[int, Set[int]]:
    if extreme not in ["min", "max"] or nr < 1:
        return -1, set()

    values = {}
    for l in lines.values():
        for x, value in enumerate(l):
            if value not in values:
                values[value] = set()
            values[value].add(x + 1)

    values = [(k, vs) for k, vs in sorted(values.items(), key=lambda i: i[0])]
    if extreme == "max":
        values.reverse()
    return values[nr - 1]


def find_extreme_lines(lines: Dict[str, List[int]], extreme: str, value: int, count: int = 1) -> Dict[int, int]:
    if extreme not in ["min", "max"] or count < 1:
        return {}

    lines_count = {l: 0 for l in lines.keys()}
    for l, values in lines.items():
        for v in values:
            if extreme == "min" and v < value or extreme == "max" and v > value:
                lines_count[l] += 1

    return {l: c for l, c in lines_count.items() if c >= count}


def get_average(lines: Dict[str, List[int]], xs: List[int], mode: str = "perX") -> Dict[object, int]:
    if mode not in ["total", "perX"]:
        return {}

    sums = {x: sum([l[x - 1] for l in lines.values()]) for x in xs}
    if mode == "perX":
        sums = {k: v / len(lines) for k, v in sums.items()}
    else:
        sums = {"total": sum(sums.values()) / (len(lines) * len(xs))}

    return sums


def calc_answers(reload_csv: bool = False) -> pd.DataFrame:
    """
    Gathers all the answers of the study participants in one dict.

    :return:
    """
    os.makedirs(TEMPORARY_OUTPUT_PATH, exist_ok=True)
    file_path = f"{TEMPORARY_OUTPUT_PATH}/S1-answer-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        df = pd.read_csv(file_path)
        df["value"] = df.apply(lambda row: data_string_converter(row["value"]), axis=1)
        return df

    # The correct result for the Y-values for B1.3 of all participants
    bt1_q3 = [[11, 11, 11, 11, 11, 9, 11, 13, 11, 13, 11, 18, 11, 13, 11, 11, 18, 9, 13, 11, -1, 11],
              [42, 45, 45, 42, 45, 45, 39, 39, 45, 42, 45, 42, 42, 45, 45, 45, 42, 45, 39, 45, -1, 42],
              [38, 46, 45, 46, 44, 46, 50, 45, 49, 44, 45, 49, 49, 46, 49, 49, 44, 46, 44, 46, -1, 44],
              [10, 10, 7, 10, 8, 9, 10, 8, 8, 8, 8, 8, 10, 10, 8, 10, 10, 8, 10, 8, -1, 8],
              [35, 31, 33, 31, 31, 31, 33, 33, 31, 31, 33, 31, 34, 34, 33, 31, 30, 34, 35, 31, -1, 25]]

    pre_output = []

    # Load the information from each answer file
    for dic in [d[0] for d in os.walk(INPUT_PATH)]:
        for fp in glob.glob(f"{dic}/AnswerLog-*.csv"):
            file_name = fp.split('\\')[-1]
            file_nr = int(file_name.split('-')[1])
            background_name = file_name.split('-')[2][1:-5]
            # Read each file

            file_df = pd.read_csv(fp, sep=';', skipinitialspace=True)

            for b in range(1, 26):
                bt = (b - 1) % 5 + 1
                if bt is 5:
                    continue

                for q in range(1, 5 if bt is not 3 else 4):
                    answer = file_df[(file_df.blockNr == b) & (file_df.questionNr == q)].answer
                    if answer.empty:
                        answer = "n.v."
                    else:
                        answer = str(answer.values[0])

                    try:
                        pre_output.append([bt, b, q, "", file_nr, background_name, float(answer)])
                    except:
                        pre_output.append([bt, b, q, "", file_nr, background_name, answer.strip()])

                    if bt is 3 and q is 2:
                        if answer != "n.v." and answer.strip() != "-":
                            if "-" not in answer:
                                min_v, max_v = "-", "-"
                                delta_v = float(answer)
                            else:
                                min_v, max_v = answer.split('-')
                                min_v, max_v = float(min_v), float(max_v)
                                delta_v = abs(float(max_v) - float(min_v))
                        else:
                            min_v, max_v, delta_v = "n.v.", "n.v.", "n.v."
                        pre_output.append([bt, b, q, "Min", file_nr, background_name, min_v])
                        pre_output.append([bt, b, q, "Max", file_nr, background_name, max_v])
                        pre_output.append([bt, b, q, "Delta", file_nr, background_name, delta_v])

    output_dict = {
        "blockType": [v[0] for v in pre_output],
        "block": [v[1] for v in pre_output],
        "complexity": [int(v[1] / 5 + 1) for v in pre_output],
        "note": [v[3] for v in pre_output],
        "questionNr": [v[2] for v in pre_output],
        "fileNr": [v[4] for v in pre_output],
        "background": [v[5] for v in pre_output],
        "value": [v[6] for v in pre_output],
    }
    output = pd.DataFrame(output_dict)

    output.to_csv(file_path)
    return output


def get_time_data() -> Dict:
    output = {}

    for dic in [d[0] for d in os.walk(INPUT_PATH)]:
        for fp in glob.glob(f"{dic}/ClientLog-*.csv"):
            file_name = fp.split('\\')[-1]
            file_nr = int(file_name.split('-')[1])
            background_name = file_name.split('-')[2][1:-5]

            with open(fp, newline="") as csv_file:
                reader = list(csv.reader(csv_file, delimiter=';'))
                reader.pop(0)
                reader.pop(0)
                while "StopApp" not in reader[0][3]:
                    index = 0
                    for i, row in enumerate(reader):
                        if i == 0:
                            continue
                        if "NextQuestion" in row[3] or "StopApp" in row[3] or "NextBlock" in row[3]:
                            index = i
                            break

                    b = int(reader[0][4])
                    while b <= 0:
                        reader.pop(0)
                        b = int(reader[0][4])
                    bt = b % 5

                    q = int(reader[0][5])

                    tmp_output = []

                    for i in range(0, index):
                        tmp_output.append(reader[0])
                        reader.pop(0)
                    tmp_output.append(reader[0])
                    output[(file_nr, background_name, bt, b, q)] = tmp_output

                    if "NextBlock" in reader[0][3]:
                        reader.pop(0)

    return output


def get_questionnaire_data():
    if os.path.isfile(QUESTIONNAIRE_INPUT_PATH):
        df = pd.read_csv(QUESTIONNAIRE_INPUT_PATH)
        return df

    return None
